<?php
defined('BASEPATH') or exit('No direct script access allowed');

class Zk_usersController extends CI_Controller
{
    public function __construct()
    {
        parent::__construct();
        $this->load->model('Zk_usersModel', 'zk_users');
        $this->load->model('UserModel', 'user');
        $this->load->helper('url');
        $this->load->library('form_validation');
        $this->output->set_content_type('application/json');
        $this->load->library('Middleware');
    }

    /**
     * Helper method to send JSON responses.
     */
    protected function jsonResponse($status, $message, $data = [], $httpCode = 200)
    {
        // If $status is not 200, use it as the HTTP response code
        $httpCode = ($status !== 200) ? $status : $httpCode;

        $this->output
            ->set_status_header($httpCode) // Set the HTTP status code
            ->set_content_type('application/json')
            ->set_output(json_encode([
                'status' => $status,
                'message' => $message,
                'data' => $data
            ]));
    }


    /**
     * List all records.
     */
    public function index()
    {
        try {
            $records = $this->zk_users->get_all_users();
            $this->jsonResponse(200, 'Records retrieved successfully', $records);
        } catch (Exception $e) {
            $this->jsonResponse(500, $e->getMessage(), [], 500);
        }
    }

    /**
     * Get a single record by ID.
     */
    public function show($id)
    {
        try {
            $record = $this->zk_users->get_user_by_id($id);
            if ($record) {
                $this->jsonResponse(200, 'Record retrieved successfully', $record);
            } else {
                $this->jsonResponse(500, 'Record not found', [], 404);
            }
        } catch (Exception $e) {
            $this->jsonResponse(500, $e->getMessage(), [], 500);
        }
    }

    public function getUsersCreated()
    {
        try {
            $records = $this->zk_users->getUsersCreated();
            $this->jsonResponse(200, 'Records retrieved successfully', $records);
        } catch (Exception $e) {
            $this->jsonResponse(500, $e->getMessage(), [], 500);
        }
    }

    /**
     * Create a new record.
     */
    public function create()
    {
        try {
            // Set validation rules
            $this->form_validation->set_rules('user_hash', 'user_hash is required', 'required');
            $this->form_validation->set_rules('name', 'name is required', 'required');
            $this->form_validation->set_rules('device_hash', 'device hash is required', 'required');
            $this->form_validation->set_rules('que_id', 'que id is required', 'required');

            if ($this->form_validation->run() == FALSE) {
                $this->jsonResponse(500, validation_errors(), [], 400);
                return;
            }

            // Prepare data
            $data = [
                'user_hash' => $this->input->post('user_hash', TRUE),
                'name' => $this->input->post('name', TRUE),
                'device_hash' => $this->input->post('device_hash', TRUE),
                'que_id' => $this->input->post('que_id', TRUE),
            ];

            // Create record
            $id = $this->zk_users->insert_user($data);
            if ($id) {
                $this->jsonResponse(200, 'Record created successfully', ['id' => $id], 201);
            } else {
                $this->jsonResponse(500, 'Failed to create record', [], 500);
            }
        } catch (Exception $e) {
            $this->jsonResponse(500, $e->getMessage(), [], 500);
        }
    }


    public function updateUserFinger()
    {
        try {
            $data = [
                'user_hash' => $this->input->post('user_hash', TRUE),
                'device_hash' => $this->input->post('device_hash', TRUE),
                'data' => $this->input->post('data', TRUE),
            ];

            if (empty($data['user_hash']) || empty($data['device_hash']) || empty($data['data'])) {
                $this->jsonResponse(500, 'user_hash, device_hash and data are required', [], 400);
                return;
            }

            $result = $this->zk_users->updateUserFinger($data);
            if ($result) {
                $this->jsonResponse(200, 'Record updated successfully', $data);
            } else {
                $this->jsonResponse(500, 'Failed to update record', [], 500);
            }
        } catch (Exception $e) {
            $this->jsonResponse(500, $e->getMessage(), [], 500);
        }
    }


    public function updateUserFace()
    {
        try {
            $data = [
                'user_hash' => $this->input->post('user_hash', TRUE),
                'device_hash' => $this->input->post('device_hash', TRUE),
                'data' => $this->input->post('data', TRUE),
            ];

            if (empty($data['user_hash']) || empty($data['device_hash']) || empty($data['data'])) {
                $this->jsonResponse(500, 'user_hash, device_hash and data are required', [], 400);
                return;
            }

            $result = $this->zk_users->updateUserFace($data);
            if ($result) {
                $this->jsonResponse(200, 'Record updated successfully', $data);
            } else {
                $this->jsonResponse(500, 'Failed to update record', [], 500);
            }
        } catch (Exception $e) {
            $this->jsonResponse(500, $e->getMessage(), [], 500);
        }
    }



    /**
     * Update an existing record.
     */
    public function update($id)
    {
        try {
            // Set validation rules
            $this->form_validation->set_rules('user_hash', 'user_hash is required', 'required');
            $this->form_validation->set_rules('name', 'name is required', 'required');


            // Add more validation rules as needed

            if ($this->form_validation->run() == FALSE) {
                $this->jsonResponse(500, validation_errors(), [], 400);
                return;
            }

            // Prepare data
            $data = [
                'user_hash' => $this->input->post('user_hash', TRUE), // XSS filtering
                'name' => $this->input->post('name', TRUE),

                // Add more fields as needed
            ];

            // Update record
            $result = $this->zk_users->update_user($id, $data);
            if ($result) {
                $this->jsonResponse(200, 'Record updated successfully', $data);
            } else {
                $this->jsonResponse(500, 'Failed to update record', [], 500);
            }
        } catch (Exception $e) {
            $this->jsonResponse(500, $e->getMessage(), [], 500);
        }
    }

    /**
     * Delete a record.
     */
    public function delete($id)
    {
        try {
            $result = $this->zk_users->delete_user($id);
            if ($result) {
                $this->jsonResponse(200, 'Record deleted successfully');
                $this->jwtmiddleware->insertMutedNotification();

            } else {
                $this->jsonResponse(500, 'Failed to delete record', [], 500);
            }
        } catch (Exception $e) {
            $this->jsonResponse(500, $e->getMessage(), [], 500);
        }
    }
}