<?php
defined('BASEPATH') or exit('No direct script access allowed');

class DocManager extends CI_Model
{

    private $storagePath;
    private $fillable_extensions = array('docx', 'xlsx', 'pptx'); // example: adjust as required
    private $viewable_extensions = array('docx', 'xlsx', 'pptx', 'pdf');
    private $editable_extensions = array('docx', 'xlsx');
    private $convertible_extensions = array('docx', 'xlsx', 'pptx');

    public function __construct()
    {
        parent::__construct();
        $this->storagePath = FCPATH . 'uploads/'; // Define a directory to store documents
        if (!is_dir($this->storagePath)) {
            mkdir($this->storagePath, 0777, true);
        }
    }

    // Generate a unique filename if it already exists
    private function getCorrectName($filename)
    {
        $path = $this->storagePath . $filename;
        $ext = pathinfo($filename, PATHINFO_EXTENSION);
        $basename = pathinfo($filename, PATHINFO_FILENAME);

        $i = 1;
        while (file_exists($path)) {
            $filename = "{$basename} ({$i}).{$ext}";
            $path = $this->storagePath . $filename;
            $i++;
        }

        return $filename;
    }

    // Create a sample document
    public function createSample($fileType, $sample, $title = null)
    {
        $ext = $this->getInternalExtension($fileType);

        // Sanitize the title to ensure it is filesystem-friendly
        $safeTitle = preg_replace('/[^A-Za-z0-9_\-]/', '', $title);

        // Use the title to generate the filename if provided; otherwise, fall back to default naming
        if (!empty($safeTitle)) {
            $filename = $this->getCorrectName($safeTitle . $ext);
        } else {
            $sampleName = ($sample == 'true') ? 'sample' : 'new';
            $filename = $this->getCorrectName($sampleName . $ext);
        }

        $filePath = $this->storagePath . $filename;

        // Copy a sample file from the assets directory (make sure the sample files exist there)
        $sampleName = ($sample == 'true') ? 'sample' : 'new';
        $sampleFile = FCPATH . "assets/document-templates/{$sampleName}{$ext}";
        if (!file_exists($sampleFile)) {
            throw new Exception("Sample file not found for type: " . $fileType);
        }

        copy($sampleFile, $filePath);

        return $filename;
    }

    /*----------------------------------------------
     * Capability Check Functions
     *---------------------------------------------*/

    public function isCanFillForms($ext)
    {
        // Remove dot if needed and check in allowed extensions
        return in_array(ltrim($ext, '.'), $this->fillable_extensions);
    }

    public function isCanView($ext)
    {
        return in_array(ltrim($ext, '.'), $this->viewable_extensions);
    }

    public function isCanEdit($ext)
    {
        return in_array(ltrim($ext, '.'), $this->editable_extensions);
    }

    public function isCanConvert($ext)
    {
        return in_array(ltrim($ext, '.'), $this->convertible_extensions);
    }

    public function isSupportedExt($ext)
    {
        return $this->isCanView($ext) ||
            $this->isCanEdit($ext) ||
            $this->isCanConvert($ext) ||
            $this->isCanFillForms($ext);
    }

    /*----------------------------------------------
     * File Extension and Template Functions
     *---------------------------------------------*/

    public function getInternalExtension($fileType)
    {
        $mapping = array(
            'word' => '.docx',
            'cell' => '.xlsx',
            'slide' => '.pptx',
            'pdf' => '.pdf'
        );
        return isset($mapping[$fileType]) ? $mapping[$fileType] : '.docx';
    }

    public function getTemplateImageUrl($fileType)
    {
        $path = $this->getServerUrl(false) . 'static/images/';
        $mapping = array(
            'word' => $path . 'file_docx.svg',
            'cell' => $path . 'file_xlsx.svg',
            'slide' => $path . 'file_pptx.svg'
        );
        return isset($mapping[$fileType]) ? $mapping[$fileType] : $path . 'file_docx.svg';
    }

    /*----------------------------------------------
     * URL Generation Functions
     *---------------------------------------------*/

    public function getServerUrl($forDocumentServer = false)
    {
        // If you have a configured document server URL, use it.
        // Otherwise, fall back to base_url() (make sure base_url is set in your config)
        if ($forDocumentServer) {
            $docServerUrl = $this->config->item('document_server_url');
            if ($docServerUrl) {
                return $docServerUrl;
            }
        }
        return base_url('/uploads/');
    }

    public function getFileUri($filename, $forDocumentServer = false)
    {
        $host = $this->getServerUrl($forDocumentServer);
        $userAddress = $this->input->ip_address();
        // Assuming you have a STATIC_URL or similar setup; adjust as necessary
        $staticUrl = base_url('static/');
        return $host . $staticUrl . $userAddress . '/' . $filename;
    }

    public function getCallbackUrl($filename)
    {
        $host = $this->getServerUrl(true);
        $userAddress = $this->input->ip_address();
        return $host . 'track?filename=' . urlencode($filename) . '&userAddress=' . urlencode($userAddress);
    }

    public function getCreateUrl($fileType)
    {
        $host = $this->getServerUrl(false);
        return $host . 'create?fileType=' . urlencode($fileType);
    }

    public function getDownloadUrl($filename, $isServerUrl = true)
    {
        $host = $this->getServerUrl($isServerUrl);
        return "$host$filename";

    }

    /*----------------------------------------------
     * File System Functions
     *---------------------------------------------*/

    public function getRootFolder()
    {
        $directory = $this->storagePath . '/';
        if (!is_dir($directory)) {
            mkdir($directory, 0777, true);
        }
        return $directory;
    }

    public function getStoragePath($filename)
    {
        return $this->getRootFolder() . $filename;
    }

    public function getHistoryPath($filename, $file, $version)
    {
        $rootFolder = $this->getRootFolder();
        $historyDir = $rootFolder . $filename . '-hist/' . $version . '/';
        return $historyDir . $file;
    }

    public function getForcesavePath($filename, $create = false)
    {
        $rootFolder = $this->getRootFolder();
        $historyDir = $rootFolder . $filename . '-hist/';
        if (!is_dir($historyDir)) {
            if ($create) {
                mkdir($historyDir, 0777, true);
            } else {
                return "";
            }
        }
        $filePath = $historyDir . $filename;
        if (!file_exists($filePath) && !$create) {
            return "";
        }
        return $filePath;
    }

    public function getStoredFiles()
    {
        $rootFolder = $this->getRootFolder();
        $files = array();

        if ($handle = opendir($rootFolder)) {
            while (false !== ($entry = readdir($handle))) {
                if ($entry != "." && $entry != ".." && is_file($rootFolder . $entry)) {
                    $files[] = $entry;
                }
            }
            closedir($handle);
        }
        // Optionally sort files by modification time (newest first)
        usort($files, function ($a, $b) use ($rootFolder) {
            return filemtime($rootFolder . $b) - filemtime($rootFolder . $a);
        });

        $fileInfos = array();
        foreach ($files as $f) {
            $ext = pathinfo($f, PATHINFO_EXTENSION);
            $fileInfos[] = array(
                'isFillFormDoc' => $this->isCanFillForms('.' . $ext),
                'version' => '1.0', // You might implement a versioning system here
                'type' => $this->getFileType($f),
                'title' => $f,
                'url' => $this->getFileUri($f, true),
                'canEdit' => $this->isCanEdit('.' . $ext)
            );
        }
        return $fileInfos;
    }

    // Simple helper to get file type from extension
    public function getFileType($filename)
    {
        $ext = strtolower(pathinfo($filename, PATHINFO_EXTENSION));
        $mapping = array(
            'docx' => 'word',
            'xlsx' => 'cell',
            'pptx' => 'slide',
            'pdf' => 'pdf'
        );
        return isset($mapping[$ext]) ? $mapping[$ext] : 'word';
    }

    public function getFileExtenstion($filename)
    {
        return pathinfo($filename, PATHINFO_EXTENSION);
    }

    /*----------------------------------------------
     * File Operation Functions
     *---------------------------------------------*/

    // Create (or copy) a file from a source path to a destination.
    public function createFile($sourcePath, $destPath, $meta = false)
    {
        if (!copy($sourcePath, $destPath)) {
            throw new Exception("Failed to copy file from $sourcePath to $destPath");
        }
        if ($meta) {
            // Implement metadata creation if needed (e.g., storing author, creation time, etc.)
        }
    }

    // Save file contents to a path
    public function saveFile($content, $path)
    {
        if (file_put_contents($path, $content) === false) {
            throw new Exception("Failed to save file to $path");
        }
    }

    // Download a file from a given URI (using cURL)
    public function downloadFileFromUri($uri, $path = null, $withSave = false)
    {
        $ch = curl_init($uri);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        $data = curl_exec($ch);
        $status = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        curl_close($ch);
        if ($status != 200) {
            throw new Exception("Document editing service returned status: $status");
        }
        if ($withSave) {
            if ($path === null) {
                throw new Exception("Path for saving file is null");
            }
            $this->saveFile($data, $path);
        }
        return $data;
    }

    // Remove a file and its history folder if it exists
    public function removeFile($filename)
    {
        $filePath = $this->getStoragePath($filename);
        if (file_exists($filePath)) {
            unlink($filePath);
        }
        $historyDir = dirname($filePath) . '/' . $filename . '-hist';
        if (is_dir($historyDir)) {
            $this->deleteDirectory($historyDir);
        }
    }

    // Remove the user folder and all its files
    public function removeUserFolder()
    {
        $rootFolder = $this->getRootFolder();
        if (strpos(realpath($rootFolder), realpath($this->storagePath)) !== 0) {
            throw new Exception("Invalid directory");
        }
        $this->deleteDirectory($rootFolder);
    }

    // Helper to recursively delete a directory
    private function deleteDirectory($dir)
    {
        if (!is_dir($dir))
            return;
        $files = array_diff(scandir($dir), array('.', '..'));
        foreach ($files as $file) {
            (is_dir("$dir/$file")) ? $this->deleteDirectory("$dir/$file") : unlink("$dir/$file");
        }
        rmdir($dir);
    }

    /*----------------------------------------------
     * Key Generation and File Info Functions
     *---------------------------------------------*/

    public function generateFileKey($filename)
    {
        $filePath = $this->getStoragePath($filename);
        $uri = $this->getFileUri($filename, false);
        if (!file_exists($filePath)) {
            return '';
        }
        $mtime = filemtime($filePath);
        // Create a hash from the URI and modification time
        $hash = hash('sha256', $uri . '_' . $mtime);
        // Replace any non-alphanumeric characters and return the first 20 characters
        $key = substr(preg_replace('/[^0-9a-zA-Z=_-]/', '_', $hash), 0, 20);
        return $key;
    }

    public function generateRevisionId($expectedKey)
    {
        if (strlen($expectedKey) > 20) {
            $expectedKey = hash('sha256', $expectedKey);
        }
        return substr(preg_replace('/[^0-9a-zA-Z=_-]/', '_', $expectedKey), 0, 20);
    }

    public function getFilesInfo($fileId = null)
    {
        $files = $this->getStoredFiles();
        $result = array();
        foreach ($files as $f) {
            $filePath = $this->getStoragePath($f['title']);
            $stats = stat($filePath);
            $fileInfo = array(
                "version" => '1.0', // Implement versioning if needed
                "id" => $this->generateFileKey($f['title']),
                "contentLength" => number_format($stats['size'] / 1024, 2) . " KB",
                "pureContentLength" => $stats['size'],
                "title" => $f['title'],
                "updated" => date("Y-m-d\TH:i:sO", $stats['mtime'])
            );
            if ($fileId && $fileId == $fileInfo["id"]) {
                return $fileInfo;
            }
            $result[] = $fileInfo;
        }
        return $fileId ? "File not found" : $result;
    }

    // Serve a file for download by setting proper headers
    public function download($filePath)
    {
        if (!file_exists($filePath)) {
            show_404();
        }
        $this->load->helper('download');
        $data = file_get_contents($filePath);
        $name = basename($filePath);
        force_download($name, $data);
    }


}
?>