<?php
/**
 * (c) Copyright Ascensio System SIA 2024
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * Model representing a user in the system.
 */
class DocumentUsers extends CI_Model
{
    private $users = [];
    private $defaultUser;

    public function __construct()
    {
        parent::__construct();

        // Predefined user descriptions
        $descr_user_1 = [
            "File author by default",
            "Doesn’t belong to any group",
            "Can review all the changes",
            "Can perform all actions with comments",
            "The file favorite state is undefined",
            "Can create files from templates using data from the editor",
            "Can see the information about all users",
            "Has an avatar",
            "Can submit forms"
        ];

        $descr_user_2 = [
            "Belongs to Group2",
            "Can review only his own changes or changes made by users with no group",
            "Can view comments, edit his own comments and comments left by users with no group. Can remove his own comments only",
            "This file is marked as favorite",
            "Can create new files from the editor",
            "Can see the information about users from Group2 and users who don’t belong to any group",
            "Has an avatar",
            "Can’t submit forms"
        ];

        $descr_user_3 = [
            "Belongs to Group3",
            "Can review changes made by Group2 users",
            "Can view comments left by Group2 and Group3 users. Can edit comments left by the Group2 users",
            "This file isn’t marked as favorite",
            "Can’t copy data from the file to clipboard",
            "Can’t download the file",
            "Can’t print the file",
            "Can create new files from the editor",
            "Can see the information about Group2 users",
            "Can’t submit forms",
            "Can't close history",
            "Can't restore the file version"
        ];

        $descr_user_0 = [
            "The name is requested when the editor is opened",
            "Doesn’t belong to any group",
            "Can review all the changes",
            "Can perform all actions with comments",
            "The file favorite state is undefined",
            "Can't mention others in comments",
            "Can't create new files from the editor",
            "Can’t see anyone’s information",
            "Can't rename files from the editor",
            "Can't view chat",
            "Can't protect file",
            "View file without collaboration",
            "Can’t submit forms"
        ];

        // Initializing users array
        $this->users = [
            [
                'id' => 'uid-1',
                'name' => 'John Smith',
                'email' => 'smith@example.com',
                'group' => '',
                'reviewGroups' => null,
                'commentGroups' => [],
                'userInfoGroups' => null,
                'favorite' => true,
                'deniedPermissions' => [],
                'descriptions' => $descr_user_1,
                'templates' => true,
                'avatar' => true,
                'goback' => ['blank' => false]
            ],
            [
                'id' => 'uid-2',
                'name' => 'Mark Pottato',
                'email' => 'pottato@example.com',
                'group' => 'group-2',
                'reviewGroups' => ['group-2', ''],
                'commentGroups' => [
                    'view' => "",
                    'edit' => ["group-2", ""],
                    'remove' => ["group-2"]
                ],
                'userInfoGroups' => ['group-2', ''],
                'favorite' => true,
                'deniedPermissions' => [],
                'descriptions' => $descr_user_2,
                'templates' => false,
                'avatar' => true,
                'goback' => ['text' => "Go to Documents"]
            ],
            [
                'id' => 'uid-3',
                'name' => 'Hamish Mitchell',
                'email' => null,
                'group' => 'group-3',
                'reviewGroups' => ['group-2'],
                'commentGroups' => [
                    'view' => ["group-3", "group-2"],
                    'edit' => ["group-2"],
                    'remove' => []
                ],
                'userInfoGroups' => ['group-2'],
                'favorite' => false,
                'deniedPermissions' => ["copy", "download", "print"],
                'descriptions' => $descr_user_3,
                'templates' => false,
                'avatar' => false,
                'goback' => null
            ],
            [
                'id' => 'uid-0',
                'name' => null,
                'email' => null,
                'group' => '',
                'reviewGroups' => null,
                'commentGroups' => [],
                'userInfoGroups' => [],
                'favorite' => null,
                'deniedPermissions' => ["protect"],
                'descriptions' => $descr_user_0,
                'templates' => false,
                'avatar' => false,
                'goback' => null
            ]
        ];

        // Default user
        $this->defaultUser = $this->users[0];
    }

    /**
     * Get all users.
     *
     * @return array Array of user data.
     */
    public function getAllUsers()
    {
        return $this->users;
    }

    /**
     * Find a user by ID.
     *
     * @param string|null $userId
     * @return array User data
     */
    public function getUserById($userId)
    {
        foreach ($this->users as $user) {
            if ($user['id'] === $userId) {
                return $user;
            }
        }
        return $this->defaultUser;
    }

    /**
     * Get users for mentions.
     *
     * Returns users (except the current user) with a non-null name and email.
     *
     * @param string $userId The ID of the current user.
     * @return array Array of users for mentions.
     */
    public function getUsersForMentions($userId)
    {
        return array_values(array_filter($this->users, function ($user) use ($userId) {
            return $user['id'] !== $userId && !empty($user['name']) && !empty($user['email']);
        }));
    }

    /**
     * Get users for protection feature.
     *
     * @param string $userId The ID of the current user.
     * @return array Array of users for protection.
     */
    public function getUsersForProtect($userId)
    {
        return array_values(array_filter($this->users, function ($user) use ($userId) {
            return $user['id'] !== $userId && !empty($user['name']);
        }));
    }
}
?>